#include <cstdio>
#include <cstring>
#include <cstdlib>
#include <windows.h>
#include <vector>
#include "distorm.h"
using namespace std;


int SeekRTA(FILE *f, BYTE *data, SIZE_T size, DWORD addr)
{
  int rta_count = 0;
  int sz = (int)size;
  int i;

  for(i = 0; i < sz; i++)
  {
    bool found = false;

    if(data[i] == 0xC3) found = true;
    else if(data[i] == 0xC2 && i+2 < sz) found = true;
    else if(i+2 < sz && data[i] == 0x58 && data[i+1] == 0xFF && data[i+2] == 0xE0) found = true;
    else if(i+2 < sz && data[i] == 0x5B && data[i+1] == 0xFF && data[i+2] == 0xE3) found = true;
    else if(i+2 < sz && data[i] == 0x59 && data[i+1] == 0xFF && data[i+2] == 0xE1) found = true;
    else if(i+2 < sz && data[i] == 0x5A && data[i+1] == 0xFF && data[i+2] == 0xE2) found = true;
    else if(i+2 < sz && data[i] == 0x5F && data[i+1] == 0xFF && data[i+2] == 0xE7) found = true;
    else if(i+2 < sz && data[i] == 0x5E && data[i+1] == 0xFF && data[i+2] == 0xE6) found = true;
    else if(i+2 < sz && data[i] == 0x5D && data[i+1] == 0xFF && data[i+2] == 0xE5) found = true;

    if(!found)
      continue;

    int ptr = i - 10;
    if(ptr < 0) ptr = 0;

    for(; ptr < i; ptr++)
    {
      _DecodeResult res;
      _DecodedInst  inst[16];
      unsigned int ret, k;

      res = distorm_decode(addr + (DWORD)ptr, data + ptr, i - ptr + (data[i]==0xC3 ? 1 : 3), Decode32Bits, inst, 16, &ret);

      if(strcmp((char*)inst[ret-1].mnemonic.p, "RET") != 0)
        continue;

      fprintf(f, "---- RTA @ %.8x ---- (%i)\n", (unsigned int)inst[0].offset, i - ptr + (data[i]==0xC3 ? 1 : 3));

      for(k = 0; k < ret; k++)
      {
        fprintf(f, " [%.8x] %-20s %s %s\n",
            (unsigned int)inst[k].offset,
            inst[k].instructionHex.p,
            inst[k].mnemonic.p,
            inst[k].operands.p);

      }
      fprintf(f, "\n");

      rta_count++;
    }      
  }

  return rta_count;
}

int
main(int argc, char **argv)
{
  puts("RTA finder by gynvael.coldwind//vx");

  if(argc != 2)
  {
    puts("usage: rta_finder <PID>");
    return 0;
  }

  int pid = 0;
  sscanf(argv[1], "%i", &pid);

  HANDLE hproc = OpenProcess(PROCESS_QUERY_INFORMATION | PROCESS_VM_READ, FALSE, (DWORD)pid);
  if(!hproc)
  {
    printf("Could not open process %i\n", pid);
    return 1;
  }


  SIZE_T ret;
  MEMORY_BASIC_INFORMATION meminfo;

  vector<BYTE> data;

  DWORD i;
  char buf[256];
  sprintf(buf, "RTAList_%i.txt", pid);

  FILE *f = fopen(buf, "w");
  if(!f)
  {
    printf("Could not create file %s\n", buf);
    CloseHandle(hproc);
    return 2;
  }

  int rta_total = 0;

  for(i = 0; i < 0x80000000;)
  {
    ret = VirtualQueryEx(hproc, (LPCVOID)i, &meminfo, sizeof(meminfo));

    if(ret == sizeof(meminfo))
    {
      if(meminfo.State == MEM_COMMIT &&
         ((meminfo.Protect & PAGE_EXECUTE) ||
          (meminfo.Protect & PAGE_EXECUTE_READ) ||
          (meminfo.Protect & PAGE_EXECUTE_READWRITE) ||
          (meminfo.Protect & PAGE_EXECUTE_WRITECOPY))
         )
      {
        printf("%.8x -> %.8x (%.8x) - %s: ", meminfo.BaseAddress, (DWORD)meminfo.BaseAddress + meminfo.RegionSize, meminfo.RegionSize,
               (meminfo.Protect & PAGE_EXECUTE) ? "..X" :
               (meminfo.Protect & PAGE_EXECUTE_READ) ? "R.X" :
               (meminfo.Protect & PAGE_EXECUTE_READWRITE) ? "RWX" :  "RCX");

        if(data.size() < meminfo.RegionSize)
          data.resize(meminfo.RegionSize);

        if(!ReadProcessMemory(hproc, meminfo.BaseAddress, &data[0], meminfo.RegionSize, &ret))
        {
          puts("could not copy memory");
          continue;
        }

        if(ret != meminfo.RegionSize)
        {
          puts("could not read the whole secion");
          continue;
        }

        int rta_count = SeekRTA(f, &data[0], data.size(), (DWORD)meminfo.BaseAddress);
        rta_total += rta_count;

        printf("%i rta sequences found\n", rta_count);
      }

      i += meminfo.RegionSize;

      continue;
    }

    i+=0x1000;
  }
  
  fclose(f);
  CloseHandle(hproc);

  printf("Done! Total %i rta sequences found and saved to %s!\n", rta_total, buf);

  return 0;
}

